
//////////////////////////////////////////////////////////////////
//                                                              //
//           PLINK (c) 2005-2006 Shaun Purcell                  //
//                                                              //
// This file is distributed under the GNU General Public        //
// License, Version 2.  Please see the file COPYING for more    //
// details                                                      //
//                                                              //
//////////////////////////////////////////////////////////////////

#include <iostream>
#include <iomanip>
#include <cmath>

#include "linear.h"
#include "logistic.h"
#include "helper.h"
#include "plink.h"
#include "options.h"
#include "crandom.h"
#include "sets.h"
#include "perm.h"


// Usage of Model

// Fit model                       LinearModel lm;
// Give pointer to PLINK           lm.setPlink(this);
// Set missing data                lm.setMissing();
// X/haploid?                      lm.setX(), lm.setHaploid();
// Set dependent (adds intercept)  lm.setDependent(Y);
// Addive effects, labels          lm.addAdditiveSNP(CSNP*); 
//                                 lm.label.push_back("ADD");
//	                           lm.addDominanceSNP(CSNP*);
// Sex effect?                     lm.addSexEffect();
//                                 lm.label.push_back("SEX");
// Covariates?                     lm.addCovariate(int);
//              	           lm.label.push_back("COV"+int2str(c+1));
// Interactions?       	           lm.addInteraction(int,int);
// Build design matrix             lm.buildDesignMatrix();
// Fit logistic model              lm.fitLM();
// Fit okay?                       lm.validParameters();
// Display results                 lm.displayResults();
// Get statistic                   lm.getStatistic();


vector_t Plink::glmAssoc(bool print_results, Perm & perm)
{

  if ( par::SNP_major )
    SNP2Ind();
  
  vector<double> results(nl_all);

  ofstream ASC;
  if (print_results)  
    {
      string f = par::output_file_name;
      if ( par::bt)
	{
	  f += ".assoc.logistic";
	  printLOG("Writing logistic model association results to [ " + f + " ] \n");
	}
      else
	{
	  f += ".assoc.linear";
	  printLOG("Writing linear model association results to [ " + f + " ] \n");
	}

      ASC.open(f.c_str(),ios::out);
      ASC << setw(4) << "CHR" << " " 
	  << setw(par::pp_maxsnp) << "SNP" << " " 
	  << setw(10) << "TEST" << " "
	  << setw(8) << "NMISS" << " " 
	  << setw(10) << "BETA" << " ";
      
      if (par::display_ci)
	ASC << setw(10) << string("L"+dbl2str(par::ci_level*100)) << " "
	    << setw(10) << string("U"+dbl2str(par::ci_level*100)) << " ";
      
      ASC << setw(12) << "STAT" << " " 
	  << setw(12) << "P" << " " 	  
	  << "\n";
      ASC.precision(4);
    }
  
  
      

  
  ////////////////////////////
  // Iterate over each locus
  
  for (int l=0; l<nl_all; l++)
    {	
      
      // Skip possibly
      if (par::adaptive_perm && !perm.test[l])
	continue;
      

      //////////////////////////////////////////////////////////
      // X-chromosome, haploid?
      // xchr_model 0: skip non-autosomal SNPs

      bool X=false;
      
      if ( par::xchr_model == 0 )
	{
	  if ( par::chr_sex[locus[l]->chr] ||
	       par::chr_haploid[locus[l]->chr] )
	    continue;
	}
      else 
	if (par::chr_sex[locus[l]->chr]) 
	  X=true;

      
      //////////////////////////////////////////////////////////
      // A new GLM
      
      Model * lm;

       
      //////////////////////////////////////////////////////////
      // Linear or logistic?

      if (par::bt)
	{
	  LogisticModel * m = new LogisticModel(this);
	  lm = m;
	}
      else
	{
	  LinearModel * m = new LinearModel(this);
	  lm = m;
	}
      

      //////////////////////////////////////////////////////////
      // Set missing data

      lm->setMissing();

      
      //////////////////////////////////////////////////////////
      // Record what is what (for interactions and tests
      
      int conditioning = 0;
      int sex = 0;
      int covariates = 0;


      //////////////////////////////////////////////////////////
      // Main effect of SNP

      lm->addAdditiveSNP(l); 
      lm->label.push_back("ADD");


      //////////////////////////////////////////////////////////
      // Dominance?

      if (par::twoDFmodel && ! par::chr_haploid[locus[l]->chr] )
	{
	  lm->addDominanceSNP(l);
	  lm->label.push_back("DOMDEV");
	}



      //////////////////////////////////////////////////////////
      // Conditioning SNPs?
      // (might be X or autosomal, dealth with automatically)

      if (par::conditioning_snps)
	{
	  for (int c=0; c<conditioner.size(); c++)
	    {
	      lm->addAdditiveSNP(conditioner[c]); 
	      lm->label.push_back("CSNP"+int2str(c+1));
	    }
	}
      

      //////////////////////////////////////////////////////////      
      // Sex-covariate (necessary for X chromosome models)

      if ( par::glm_sex_effect || X )
	{
	  lm->addSexEffect();
	  lm->label.push_back("SEX");	  
	}


      //////////////////////////////////////////////////////////
      // Covariates?

      if (par::clist)
	{
	  for (int c=0; c<par::clist_number; c++)
	    {
	      lm->addCovariate(c);
	      lm->label.push_back("COV"+int2str(c+1));
	    }
	}


      //////////////////////////////////////////////////////////
      // Interactions

      // addInteraction() takes parameter numbers
      // i.e. not covariate codes
      
      // 0 intercept
      // 1 A
      //   {D}
      //   {conditioning SNPs}
      //   {sex efffect}
      //   {covariates}

      // Allow for interactions between conditioning SNPs, sex, covariates, etc
      	      
      ////////////////////////////////////////
      // Basic SNP x covariate interaction? 

      if (par::simple_interaction)
	{
	  
	  // A (and D?) by conditioning SNPs, sex, covariates

	  int cindex = 2;
	  if ( par::twoDFmodel )
	    cindex = 3;
	  
	  for (int c=0; c<conditioner.size(); c++)
	    {
	      lm->addInteraction(1,cindex);
	      lm->label.push_back("ADDxCSNP"+int2str(c+1));	  
	      
	      if ( par::twoDFmodel )
		{
		  lm->addInteraction(2,cindex);
		  lm->label.push_back("DOMDEVxCSNP"+int2str(c+1));	  
		}
	      
	      cindex++;
	    }

	  if (par::glm_sex_effect)
	    {
	      lm->addInteraction(1,cindex);
	      lm->label.push_back("ADDxSEX");	  
	      
	      if ( par::twoDFmodel )
		{
		  lm->addInteraction(2,cindex);
		  lm->label.push_back("DOMDEVxSEX");
		}
	      
	      cindex++;
	    }
	  for (int c=0; c<par::clist_number; c++)
	    {
	      lm->addInteraction(1,cindex);
	      lm->label.push_back("ADDxCOV"+int2str(c+1));	  
	      
	      if ( par::twoDFmodel )
		{
		  lm->addInteraction(2,cindex);
		  lm->label.push_back("DOMDEVxCOV"+int2str(c+1));	  
		}
	      
	      cindex++;
	      
	    }
	}
      
      
      //////////////////////////////
      // Fancy X chromosome models
      
      if ( X && par::xchr_model > 2 )
	{
	  
	  // Interaction between allelic term and sex (i.e. 
	  // allow scale of male effect to vary)
	  int sindex = 2;
	  if ( par::twoDFmodel )
	    sindex++;
	  sindex += conditioner.size();
	  
	  lm->addInteraction(2,sindex);
	  lm->label.push_back("XxSEX");	  
	  
	  // xchr model 3 : test ADD + XxSEX
	  // xchr model 4 : test ADD + DOM + XxSEX
	}
      
      
      //////////////////////////////
      // Build design matrix

      lm->buildDesignMatrix();
      
      
      //////////////////////////////////////////////////
      // Fit linear or logistic model (Newton-Raphson)

      lm->fitLM();
      

      ////////////////////////////////////////
      // Check for multi-collinearity

      lm->validParameters();
      

      ////////////////////////////////////////
      // Obtain estimates and statistic

      if (print_results)
	lm->displayResults(ASC,locus[l]);


      ////////////////////////////////////////
      // Two DF model? 

      if (par::twoDFmodel)
 	{
	 
	  // Does this work with covariates?
	  
 	  vector_t h;
 	  h.resize(2,0);
 	  matrix_t H;
 	  sizeMatrix(H,2,3);

	  // Joint test of additive and dominant models
 	  H[0][0] = 0; 	  H[0][1] = 1;   H[0][2] = 0; 
	  H[1][0] = 0; 	  H[1][1] = 0;   H[1][2] = 1; 

	  double chisq = lm->isValid() ? lm->linearHypothesis(H,h) : 0;
	  double pvalue = chiprobP(chisq,2);
	  
	  ASC << setw(4) << locus[l]->chr << " " 
	      << setw(par::pp_maxsnp) << locus[l]->name << " " 
	      << setw(10) << "GENO_2DF" << " "
	      << setw(8) << lm->Ysize() << " " 
	      << setw(10) << "NA" << " ";
	  
	  if (par::display_ci)
	    ASC << setw(10) << "NA" << " "
		<< setw(10) << "NA" << " ";
	  
	  if (lm->isValid() && realnum(chisq) )
	    ASC << setw(12) << chisq << " " 
		<< setw(12) << pvalue << "\n"; 
	  else
	    ASC << setw(12) << "NA" << " " 
		<< setw(12) << "NA" << "\n"; 
	    
 	}




      /////////////////////////////////////////
      // Joint test of everything?
      
      if (par::test_full_model)
 	{
	  
	  int df = lm->getNP() - 1;
	  
   	  vector_t h;
  	  h.resize(df,0);
  	  matrix_t H;
  	  sizeMatrix(H,df,lm->getNP());
	  
 	  // Joint test of dominant and recessive
	  for (int i=0; i<df; i++)
	    {
	      // Intercept
	      H[i][0] = 0;
	      for (int j=1; j<lm->getNP(); j++)
		H[i][j] = i+1 == j ? 1 : 0;
	    }
	  
	  double chisq = lm->isValid() ? lm->linearHypothesis(H,h) : 0;
	  double pvalue = chiprobP(chisq,df);
	  
	  ASC << setw(4) << locus[l]->chr << " " 
	      << setw(par::pp_maxsnp) << locus[l]->name << " " 
	      << setw(10) << "FULL_"+int2str(df)+"DF" << " "
	      << setw(8) << lm->Ysize() << " " 
	      << setw(10) << "NA" << " ";
	  
	  if (par::display_ci)
	    ASC << setw(10) << "NA" << " "
		<< setw(10) << "NA" << " ";
	  
	  if (lm->isValid() && realnum(chisq) )
	    ASC << setw(12) << chisq << " " 
		<< setw(12) << pvalue << "\n"; 
	  else
	    ASC << setw(12) << "NA" << " " 
		<< setw(12) << "NA" << "\n"; 
	  
	}
      


      //////////////////////////////
      // Fancy X chromosome models

      if ( X && par::xchr_model == 3 )
	{
	  
	  vector_t h;
	  h.resize(2,0);
 	  matrix_t H;
 	  sizeMatrix(H,2,3);

	  // Joint test of additive and dominant models
 	  H[0][0] = 0; 	  H[0][1] = 1;   H[0][2] = 0; 
	  H[1][0] = 0; 	  H[1][1] = 0;   H[1][2] = 1; 

	  double chisq = lm->isValid() ? lm->linearHypothesis(H,h) : 0;
	  double pvalue = chiprobP(chisq,2);
	  
	  ASC << setw(4) << locus[l]->chr << " " 
	      << setw(par::pp_maxsnp) << locus[l]->name << " " 
	      << setw(10) << "XMOD_2DF" << " "
	      << setw(8) << lm->Ysize() << " " 
	      << setw(10) << "NA" << " ";
	  
	  if (par::display_ci)
	    ASC << setw(10) << "NA" << " "
		<< setw(10) << "NA" << " ";
	  
	  if (lm->isValid() && realnum(chisq) )
	    ASC << setw(12) << chisq << " " 
		<< setw(12) << pvalue << "\n"; 
	  else
	    ASC << setw(12) << "NA" << " " 
		<< setw(12) << "NA" << "\n"; 
	  
	}


      
      ////////////////////////////////////////
      // Store statistic (1 df chisq)
      // ( based on value of testParameter )

      results[l] = lm->getStatistic();


      /////////////////////////////////
      // Clear up linear model

      delete lm;

      // Next SNP
    }

  
 
  
  if (print_results)
    ASC.close();
  
  return results;

}

